/* SICStus - Perl extension to provide access to SICStus Prolog
 *
 * Version: 1.0 12-Jan-1995
 * Author:  Charles Bailey  bailey@genetics.upenn.edu
 *
 * Revision History:
 *   1.0  12-Jan-1995  Charles Bailey  bailey@genetics.upenn.edu
 *    - original version
 *
 * This work is copyright 1995 by Charles Bailey.  All rights reserved.
 * Permission is granted to use and copy this work under the same term as
 * Perl itself (See the file Copying in the Perl distribution for
 * details.)
 * 
 * NO WARRANTY is made that this software will function in any particular
 * manner or be fit for any purpose.  The author, the University of
 * Pennsylvania, and its affiliates will not be liable for any damages
 * whatsoever, including but not limited to direct, indirect, and
 *consequential damages, arising from the use of this software.
 */

/* Perl's C header files */
#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"

/* SICStus' C header files */
#include "runtime.h"   /* from $(SP_PATH)/Runtime */
#include "flinkage.c"  /* generated in $CWD by prepare_foreign_files/1 */

/* Buffer for fake I/O 'stream' used to pass strings to Prolog */
typedef struct _stmbuf {
  int idx;
  SV *cmd;
} Sbuf;

/* Record of information for open query */
typedef struct _qrydata {
  SP_qid qid;
  unsigned int arity;
  SP_term *terms;
  SV **vars;
} Qry;

static int isdefined(sv)
     SV *sv;
{
  dSP;
  SV **oldsp = stack_sp;
  int retval;

  PUSHs(sv);
  PUTBACK;

  pp_defined();
  SPAGAIN;
  retval = (*sp == &sv_yes);
  stack_sp = oldsp;

  return retval;
}

static int spgetc(buf)
     Sbuf *buf;
{
  char *str;
  STRLEN len;

  str = SvPV(buf->cmd,len);
  if (buf->idx < len) return str[buf->idx++];
  else return 0;
}

static int speof(buf)
     Sbuf *buf;
{
  return !SvPOK(buf->cmd) || (buf->idx > SvCUR(buf->cmd));
}

static int spclose(buf)
     Sbuf *buf;
{
  SvREFCNT_dec(buf->cmd);
  Safefree(buf);
  return 0;
}

static SP_stream *svtostream(sv)
     SV *sv;
{
  char *str;
  STRLEN len;
  Sbuf *buf;
  SP_stream *newstream;

  New(1702,buf,1,Sbuf);
  buf->idx = 0;
  buf->cmd = sv;
  SvREFCNT_inc(sv);

  SP_make_stream(buf,spgetc,NULL,NULL,speof,NULL,spclose,&newstream);

  return newstream;
}  /* end of svtostream() */

static SP_term svtolist(sv)
   SV *	sv;
{
  char *idx;
  STRLEN len;
  SP_term head,tail,chr;

  SP_put_string(&tail,"[]");
  SP_show_term(&tail);

  idx = SvPV(sv,len);
  while (len) {
    SP_put_compound(&head,".",2);
    SP_put_integer(&chr,idx[--len]);
    SP_put_arg(1,head,chr);
    SP_put_arg(2,head,tail);
    tail = head;
  }

  SP_hide_term(&tail);
  return tail;
}  /* end of svtolist() */


static SP_term svtoterm(sv)
     SV *sv;
{
  SP_term t;

  if (sv_isa(sv,"SICStus::Atom")) {
    SP_put_string(&t,SvPV(SvRV(sv),na));
  }
  else if (sv_isa(sv,"SICStus::Integer")) {
    SP_put_integer(&t,SvIV(SvRV(sv)));
  }
  else if (sv_isa(sv,"SICStus::Float")) {
    SP_put_float(&t,SvNV(SvRV(sv)));
  }
  else if (sv_isa(sv,"SICStus::List")) {
    if (!SvCUR(SvRV(sv))) SP_put_string(&t,"[]");
    else t = svtolist(SvRV(sv));
  }
  else if (SvNOK(sv))
    SP_put_float(&t,SvNV(sv));
  else if (SvIOK(sv))
    SP_put_integer(&t,SvIV(sv));
  else if (SvPOK(sv)) {
    if (!SvCUR(sv)) SP_put_string(&t,"[]");
    else t = svtolist(sv);
  }

  return t;
}  /* end of svtoterm() */


static SV *listtosv(term,sv)
     SP_term term;
     SV *sv;
{
  char tmp[256], *cp;
  int arity, chr, toplevel = 1;
  SP_term head, end;
  STRLEN idx = 0;
  SV *newsv;

  if (sv) sv_setpvn(sv,"",0);
  else newsv = newSVpv("",0);
  SP_put_string(&end,"[]");
  while (1) {
    if (SP_is_atom(term) && !SP_compare(term,end)) {
      if (idx) sv_catpvn(sv ? sv : newsv,tmp,idx);
      return sv ? sv : newsv;
    }
    if (!SP_is_compound(term) ||
	!SP_get_compound(term,&cp,&arity)) {
      if (sv) return (sv_setsv(sv,&sv_undef),sv);
      else {
	SvREFCNT_dec(newsv);
	return &sv_undef;
      }
    }
    if (strNE(cp,".") || arity != 2) {
      if (toplevel) {
	sprintf(tmp,"%.*s/%d",sizeof tmp - 3,cp,arity);
	sv_setpv(sv ? sv : newsv,tmp);
	return sv ? sv : newsv;
      }
      else {
	if (sv) return (sv_setsv(sv,&sv_undef),sv);
        else {
	  SvREFCNT_dec(newsv);
	  return &sv_undef;
	}
      }
    }
    if (!SP_get_arg(1,term,&head) ||
	!SP_get_integer(head,&chr) ||
	!SP_get_arg(2,term,&head)) {
	if (sv) return (sv_setsv(sv,&sv_undef),sv);
        else {
	  SvREFCNT_dec(newsv);
	  return &sv_undef;
	}
    }
    tmp[idx++] = chr;
    if (idx == sizeof tmp) {
      sv_catpvn(sv ? sv : newsv,tmp,sizeof tmp);
      idx = 0;
    }
    term = head;
    toplevel = 0;
  }

}  /* end of listtosv() */


static SV *termtosv(term,sv)
     SP_term term;
     SV *sv;
{
  char *cp;
  double realval;
  long intval;
  SV *newsv;

  switch (SP_term_type(term)) {
    case SP_TYPE_VARIABLE:
      if (sv) sv_setsv(sv,&sv_undef);
      else newsv = &sv_undef;
      break;
    case SP_TYPE_INTEGER:
      if (!SP_get_integer(term,&intval))
	croak("SP_get_integer/2: failed to get Prolog value");
      if (sv) sv_setiv(sv,intval);
      else newsv = newSViv(intval);
      break;
    case SP_TYPE_FLOAT:
      if (!SP_get_float(term,&realval))
	croak("SP_get_float/2: failed to get Prolog value");
      if (sv) sv_setnv(sv,realval);
      else newsv = newSVnv(realval);
      break;      
    case SP_TYPE_ATOM:
      if (!SP_get_string(term,&cp))
	croak("SP_get_string/2: failed to get Prolog value");
      if (sv) sv_setpv(sv,cp);
      else newsv = newSVpv(cp,0);
      break;      
    case SP_TYPE_COMPOUND:
      if (sv) listtosv(term,sv);
      else newsv = listtosv(term,NULL);
      break;
    default:
      if (sv) sv_setsv(sv,sv_undef);
      else newsv = &sv_undef;
    }

  return sv ? sv : newsv;

}  /* end of termtosv() */

static SV *new_mortalobj(pkg)
     char *pkg;
{
  HV *stash;
  SV *mysv, *myref;

  stash = gv_stashpv(pkg,FALSE);
  mysv = sv_2mortal(newSViv(0));
  myref = sv_2mortal(newRV(mysv));
  return sv_bless(myref,stash);
}  /* end of new_mortalobj() */

int pass_term_SPtoPerl(stm,name,term)
     int stm;       /* reserved for future use (multiple interpreters) */
     char *name;    /* name of Prolog term */
     SP_term term;  /* 'value' of Prolog term */
{
  SV *sv;

  if ((sv = perl_get_sv(name,TRUE)) == NULL)
    croak("SICStus:: Unable to create Perl variable");
  (void) termtosv(term,sv);
}

static SP_pred_ref
parse_predspec(spec,arity)
     char *spec;
     int arity;
{
  char *name, *module = (char *)0;
  SP_pred_ref pref;

  if ((name = strchr(spec,':')) != NULL) {
    *(name++) = '\0';
    module = spec;
  }
  else name = spec;
  pref = SP_predicate(name,arity,module);
  if (module) *(name - 1) = ':';
  if (pref == NULL) {
    croak("Can't find Prolog predicate %s/%d",spec,arity);
  }

  return pref;

}  /* end of parse_predspec() */

/* This hack brought to you by C's limited vararg processing */
static I32
vararg_call(func,pred,arity,myargv)
     I32 (*func)();
     SP_pred_ref pred;
     int arity;
     SP_term *myargv;
{
  int retval;

  switch (arity) {
    case 0:
      retval = (I32) (*func)(pred);
      break;
    case 1:
      retval = (I32) (*func)(pred,&myargv[0]);
      break;
    case 2:
      retval = (I32) (*func)(pred,&myargv[0],&myargv[1]);
      break;
    case 3:
      retval = (I32) (*func)(pred,&myargv[0],&myargv[1],&myargv[2]);
      break;
    case 4:
      retval = (I32) (*func)(pred,&myargv[0],&myargv[1],&myargv[2],&myargv[3]);
      break;
    case 5:
      retval = (I32) (*func)(pred,&myargv[0],&myargv[1],&myargv[2],
		       &myargv[3],&myargv[4]);
      break;
    case 6:
      retval = (I32) (*func)(pred,&myargv[0],&myargv[1],&myargv[2],
		       &myargv[3],&myargv[4],&myargv[5]);
      break;
    case 7:
      retval = (I32) (*func)(pred,&myargv[0],&myargv[1],&myargv[2],
		       &myargv[3],&myargv[4],&myargv[5],&myargv[6]);
      break;
    case 8:
      retval = (I32) (*func)(pred,&myargv[0],&myargv[1],&myargv[2],
		       &myargv[3],&myargv[4],&myargv[5],
		       &myargv[6],&myargv[7]);
      break;
  }

  return retval;

}  /* end of vararg_call() */


MODULE = SICStus  PACKAGE= SICStus PREFIX= SP_

int
SP_initialize()

int
SP_load(filename)
  char *	filename


void
SP_eval(self,goal)
  SV *	self
  SV *	goal
  CODE:
  static SP_pred_ref evalhook;
  char *cp;
  int sts;
  SP_term stmno,stmterm,errterm;

  if (!evalhook) {
    evalhook = SP_predicate("evalperlstr",1,"perl");
    if (!evalhook)
      croak("doSPeval: Can't connect to Prolog hook");
  }
  SP_put_address(&stmno,svtostream(goal));
  SP_put_compound(&stmterm,"$stream",1);
  SP_put_arg(1,stmterm,stmno);

  if ((sts = SP_query(evalhook,&stmterm)) == -1) {
    SP_retract_exception(&errterm);
    SP_get_compound(errterm,&cp,&sts);
    croak(cp);
  }

  ST(0) = sts ? &sv_yes : &sv_no;


void
SP_query(self,predspec,...)
  SV *	self
  char *	predspec
  CODE:
  char *errname;
  register int i, retval, arity;
  SP_pred_ref pref;
  SP_term termlist[8];

  if (items > 10)
    croak("SP_query cannot handle predicates with arity >8");
  arity = items - 2;
  pref = parse_predspec(predspec,arity);
  for (i = 0; i < arity; i++) {
    if (isdefined(ST(i+2)))
      termlist[i] = svtoterm(ST(i+2));
    else
      SP_put_variable(&termlist[i]);
  }
  retval = (int) vararg_call(SP_query,pref,arity,termlist);
  if (retval == -1) {
    SP_retract_exception(&termlist[0]);
    SP_get_compound(termlist[0],&errname,&retval);
    croak(errname);
  }
  else if (retval == 0)
    ST(0) = &sv_no;
  else {
    for (i = 0; i < arity; i++)
      if (!SvREADONLY(ST(i+2))) (void) termtosv(termlist[i],ST(i+2));
    ST(0) = &sv_yes;
  }


MODULE = SICStus  PACKAGE= SICStus::Query PREFIX= SP_


void
SP_open_query(self,predspec,...)
  SV *	self
  char *	predspec
  CODE:
  register int i, retval, arity;
  SP_pred_ref pref;
  Qry *query;

  if (items > 9)
    croak("SP_query cannot handle predicates with arity >8");
  arity = items - 2;
  pref = parse_predspec(predspec,arity);
  query = (Qry *) safemalloc(sizeof(Qry) + arity*(sizeof(SP_term) +
						  sizeof(SV *)));
  query->arity = arity;
  query->terms = (SP_term *) ((char *) query + sizeof(Qry));
  query->vars = (SV **) ((char *) (query->terms) + arity * sizeof(SP_term));
  for (i = 0; i < arity; i++) {
    query->vars[i] = ST(i+2);
    SvREFCNT_inc(query->vars[i]);
    if (isdefined(ST(i+2))) query->terms[i] = svtoterm(ST(i+2));
    else SP_put_variable(&(query->terms[i]));
    SP_show_term(&(query->terms[i]));
  }
  if ((query->qid = (SP_qid)
       vararg_call(SP_open_query,pref,arity,query->terms)) == NULL) {
    for (i = 0; i < arity; i++) {
      SvREFCNT_dec(query->vars[i]);
      SP_hide_term(&(query->terms[i]));
    }
    Safefree(query);
    croak("Unable to open SICStus query for %s/%d",predspec,arity);
  }
  ST(0) = new_mortalobj("SICStus::Query");
  sv_setiv(SvRV(ST(0)),(I32) query);


void
SP_next_solution(queryref)
  SV *	queryref
  CODE:
  char *cp;
  register int i, retval;
  Qry *query;
  SP_term errterm;

  if(!SvROK(queryref))  /* don't check package to allow subclasses */
    croak("Not a SICStus query object");
  query = (Qry *) SvIV(SvRV(queryref));
  if (!query) ST(0) = &sv_no;
  else {
    if ((retval = SP_next_solution(query->qid)) == -1) {
      SP_retract_exception(&errterm);
      SP_get_compound(errterm,&cp,&retval);
      SP_close_query(query->qid);
      for (i = 0; i < query->arity; i++) {
	SvREFCNT_dec(query->vars[i]);
	SP_hide_term(&(query->terms[i]));
      }
      Safefree(query);
      sv_setiv(SvRV(queryref),0);
      croak(cp);
    }
    else if (!retval) {
      SP_close_query(query->qid);
      for (i = 0; i < query->arity; i++) {
	SvREFCNT_dec(query->vars[i]);
	SP_hide_term(&(query->terms[i]));
      }
      Safefree(query);
      sv_setiv(SvRV(queryref),0);
      ST(0) = &sv_no;
    }    
    else {
      for (i = 0; i < query->arity; i++) {
	if (!SvREADONLY(query->vars[i]))
	  (void) termtosv(query->terms[i],query->vars[i]);
      }
      ST(0) = &sv_yes;
    }
  }

void
SP_cut_query(queryref)
  SV *	queryref
  CODE:
  char *cp;
  register int i;
  Qry *query;
  SP_term errterm;

  if(!SvROK(queryref))  /* don't check package to allow subclasses */
    croak("Not a SICStus query object");
  query = (Qry *) SvIV(SvRV(queryref));
  if (!query) ST(0) = &sv_no;
  else {
    if (!SP_cut_query(query->qid)) {
      SP_close_query(query->qid);
      for (i = 0; i < query->arity; i++) {
	SvREFCNT_dec(query->vars[i]);
	SP_hide_term(&(query->terms[i]));
      }
      Safefree(query);
      sv_setiv(SvRV(queryref),0);
      croak("SP_cut_query/1 failed");
    }
    ST(0) = &sv_yes;
  }

void
SP_close_query(queryref)
  SV *	queryref
  CODE:
  char *cp;
  register int i,retval;
  Qry *query;
  SP_term errterm;

  if(!SvROK(queryref))  /* don't check package to allow subclasses */
    croak("Not a SICStus query object");
  query = (Qry *) SvIV(SvRV(queryref));
  if (!query) ST(0) = &sv_yes;
  else {
    retval = SP_close_query(query->qid);
    for (i = 0; i < query->arity; i++) {
      SvREFCNT_dec(query->vars[i]);
      SP_hide_term(&(query->terms[i]));
    }
    Safefree(query);
    if (!retval) croak("SP_close_query/1 failed");
    else {
      sv_setiv(SvRV(queryref),0);
      ST(0) = &sv_yes;
    }
  }

