#!../../perl -I../../lib
#This work is copyright 1995 by Charles Bailey.  All rights reserved.
#Permission is granted to use and copy this work under the same term as
#Perl itself (See the file Copying in the Perl distribution for
#details.)
#
#NO WARRANTY is made that this software will function in any particular
#manner or be fit for any purpose.  The author, the University of
#Pennsylvania, and its affiliates will not be liable for any damages
#whatsoever, including but not limited to direct, indirect, and
#consequential damages, arising from the use of this software.
#
# Short demo Perl script for SICStus Prolog interface.

print <<"END";
Demo of the SICStus extension v1.0 from Perl5

This demo will perform 16 tests of the SICStus extension to Perl
5.000.  While not exhaustive, these tests should provide an
illustration of the extension's basic function.  For more information,
see the code and comments in this file ($0).

END

# Pull in the SICStus code
    print "Loading SICStus extension . . .";
require SICStus;
    print "ok 1\n"; # If the require failed, Perl's quit already

# If running from a Perl interpreter, initialize the SICStus runtime system
    print "Initializing SICStus runtime system . . .";
$sts = SICStus::rtinit();
    $sts ? print "ok 2\n" : die "failed 2\n";

# After initialization, we can load Prolog ql files via the
# SICStus::load() function, which takes one arg: the file.

# We'll use the object-oriented approach here
$sp = new SICStus;

# OK, let's get some facts to Prolog
    print "Asserting clauses into Prolog . . .";
# the direct way
$sts = $sp->eval('assert(cuts("EcoRI","GAATTC")).');
    $sts ? print "ok 3 . . ." : die "failed 3";
#or the shortcut
$sts = $sp->assert(q{cuts("HinDIII","AAGCTT")});
    $sts ? print "ok 4 . . ." : die "failed 4";
$sts = $sp->assert('cuts("BamHI","GGATCC")');
    $sts ? print "ok 5 . . ." : die "failed 5";
$sts = $sp->assert(q{cuts('BstI',"GGATCC")});
    $sts ? print "ok 6\n" : die "failed 6";

# Now, see what we can get by evaling a string
# Any Prolog variables in the string which get bound
# when the goal is satisfied appear in Perl as new
# scalar variables, and SICStus::eval() returns 1.
     print "Asking what cuts at GAATTC (string eval) . . .";
$sts = $sp->eval('cuts(What,"GAATTC").');
     print "Prolog says \"$What\". . .";
     ($sts && ($What eq 'EcoRI')) ? print "ok 7\n" : die "failed 7";

# If the goal isn't satisfied, no Perl variables are created,
# and SICStus::eval() returns 0.
     print "Asking what cuts at CCGCGG (string eval) . . .";
$sts = $sp->eval('cuts(Which,"CCGCGG").');
     print "Prolog says \"$Which\". . .";
     (!$sts && ($Which eq '')) ? print "ok 8\n" : die "failed 8";

# Alternatively, we can look up a specific predicate, and
# pass it a list of arguments (the arity of the predicate
# is inferred from the number of arguments passed to the
# SICStus::query() function.  If a given argument is a
# defined Perl variable, its value gets passed to Prolog;
# if it's undefined, Prolog sees a variable, and, if the
# goal is satisfied, the Perl variable is set to the
# value of the instantiated Prolog term.
# Return value is 1 if query succeeds, 0 if it fails.

    print "Asking where HinDIII cuts (list query) . . .";
$sts = $sp->query("cuts","HinDIII",$seq);
    print "Prolog says \"$seq\" . . .";
    ($sts && ($seq eq 'AAGCTT')) ? print "ok 9\n" : die "failed 9";

# Defined Perl scalars are passed to Prolog as terms whose type is
# determined by the most recent Perl usage.  If the Perl scalar was
# last used as a number (e.g. for arithmetic), it goes to Prolog
# as a float; if it was last used as an integer (rare except if you
# specifically set it that way in C code), it goes to Prolog as an
# integer, and if it was last used as a string, it goes to Prolog
# as a character list.
# If you want to coerce a value into a certain Prolog type, you
# can 'cast' it as follows
#     $tag = new SICStus::Atom $val      # Perl string -> Prolog atom
#     $tag = new SICStus::Float $val     # Perl number -> Prolog float
#     $tag = new SICStus::Integer $val   # Perl number -> Prolog integer
#     $tag = new SICStus::List $val      # Perl string -> Prolog list
# and then use $tag in place of $var in the call to Prolog.
# When a Perl scalar is set from a Prolog term, it's initial type is
# the same as the Prolog term (both character lists and atoms appear
# as Perl strings), but since Perl interconverts types automatically
# this isn't a big deal.  One aside: At the moment, a Prolog compound
# term whose first functor isn't "." (i.e. it isn't a string) comes
# to Perl as a string representation of the functor name.  If a
# functor other than "." shows up in the middle of a compound
# term which started out as a string, the entire compound term is
# discarded, and the Perl scalar is set to undef.  This may improve
# in the future, time permitting.

    print "Asking where BstI cuts . . .";
$enz = new SICStus::Atom "BstI"; # We asserted BstI as an atom above
undef $seq;                      # clear out $seq from the last call
$sts = $sp->query("cuts",$enz,$seq);
    print "Prolog says \"$seq\" . . .";
    ($sts && ($seq eq "GGATCC")) ? print "ok 10\n" : die "failed 10";

# We're done with the SICStus object so undef it; we can also
# just let it go out of scope or ignore it
undef $sp;

# The SICStus::Query class is used to set up and repeatedly satisfy
# a Prolog query using a known predicate and list of arguments.  The
# same rules apply to input arguments as for SICStus::query().  The
# query is set up using the new() method, and returns a query handle.

    print "Asking Prolog for everything it knows about enzymes . . .\n";
    print "\topening query . . .";
undef $enz;
undef $seq;
$qry = new SICStus::Query ('cuts',$enz,$seq);
    print +($qry ? '' : 'not '),"ok 11\n";

# Once the query has been opened, an attempt is made to satisfy it
# each time the next_solution() method is called.  If this succeeds,
# it returns 1, and the Perl scalars corresponding to Prolog variables
# in the input are set to the values to which the Prolog terms were
# instantiated.  If the query fails, next_solution() returns 0, and
# no changes are made to the Perl scalars.

$test = 12;
while ($qry->next_solution()) {
     print "\tProlog says $enz cuts at $seq . . . ok $test\n";
# We can do anything we like to the Perl scalars between
# queries; the next successful solution unifies them with
# the Prolog terms again.
     $enz = 'somestring';
     $seq = 'anotherstring';
     $test++;
}
    if ($test != 16) { die "failed $test"; }
# The solutions of a query can be limited by calling the cut_query()
# method at any time; this eliminates all previously satisfied
# solutions.

# When you are finished with a query, you can close it in one of
# three ways:
#    1. When next_solution() fails (returns 0), the query is
#       automatically closed.
#    2. Call the close_query() method.
#    3. Undef the query handle.

# This is really redundant, since the failed next_solution() above
# will have closed the query.
    print "\tclosing query . . .";
$sts = $qry->close_query();
    $sts ? print "ok 16\n" : die "failed 16";

print "\nDemo completed.\n";
